;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : tmimage.scm
;; DESCRIPTION : convert texmacs fragment (selection) to image formats.
;;               Try embedding source code in image
;; COPYRIGHT   : (C) 2012-2018  Philippe Joyez
;;
;; This software falls under the GNU general public license version 3 or later.
;; It comes WITHOUT ANY WARRANTY WHATSOEVER. For details, see the file LICENSE
;; in the root directory or <http://www.gnu.org/licenses/gpl-3.0.html>.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; convert active selection to various graphics format
;; If target is svg embed texmacs code of the selection in the image for
;; re-edition (Could be done for other formats too).
;; the svg produced by this method can be pasted in inkscape via the clipboard
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(texmacs-module (convert images tmimage)
  (:use (convert tmml tmmlout)
        (convert tmml tmtmml)))

;; (display "Texmacs] Loading module tmimage\n")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Handling of image convertion preferences
;; And preference widget 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-preferences
  ("texmacs->image:format" "svg" noop)
  ("texmacs->image:scale" "2.0" noop)
  ("texmacs->image:raster-resolution" "300" noop)
  ("equation-editor-server" "off" noop))

(define (pretty-format-list)
  (let* ((desired-image-format-list '(("svg" "Svg")  ("eps" "Eps")
           ("png" "Png")("tif" "Tiff") ("jpg" "Jpeg")))
         (valid-image-format-list 
           (filter (lambda (x) (file-converter-exists? "x.pdf" (string-append "y." (car x)) ))
             desired-image-format-list)))
   (eval `(define-preference-names "texmacs->image:format" ,@ valid-image-format-list))
   (cadr (apply map list valid-image-format-list))))

(define (svg-converter-available)
  (nnot (converter-search "pdf-file" "svg-file")))

(define (inkscape-detected)
  (if (or (os-mingw?) (os-win32?)) #t (url-exists-in-path? "inkscape"))) ;;no reliable way on windows

(define (inkscape-prefs-dir)
  (url-concretize (url-append (if (or (os-win32?) (os-mingw?))
  (system->url "$APPDATA") (system->url
  "~/.config"))(string->url "inkscape"))))
  
(define (inkscape-extension)
  (url-append (inkscape-prefs-dir) (string->url "extensions/texmacs.inx")))

(define (inkscape-prefs-found)
  (url-exists? (inkscape-prefs-dir)))

(define (inkscape-extension-installed)
  (url-exists? (inkscape-extension)))

(define (get-svg-converter)
  (if (or (os-win32?) (os-mingw?)) 
    (show-message "pdftocairo is bundled with TeXmacs at this address: https://github.com/slowphil/mingw-w64-texmacs/releases"
      "'pdftocairo' is needed for co-operating with Inkscape, but it is not present.")
    (if (os-macos?) 
      (show-message "Please install either 'pdf2svg' or 'pdftocairo' (included in poppler) from MacPorts, Fink or Homebrew"
        "missing svg converter")
      (show-message "Please install either 'pdf2svg' or 'pdftocairo' (included in poppler-utils) using your distribution package management system"
        "missing svg converter"))))
(define (start-inkscape)
  (if (or (os-win32?) (os-mingw?))
    (show-message "Make sure Inkscape is installed on your system, and then start it once." "missing Inkscape preferences")
    (shell "inkscape")))

(define (install-inkscape-extension)
  (let ((dest-dir (escape-shell (url-concretize 
          (url-append 
            (if (or (os-win32?) (os-mingw?)) (system->url "$APPDATA")  (system->url "~/.config"))
            (string->url "inkscape/extensions")))))
         (source-dir (escape-shell (url-concretize (url-unix "$TEXMACS_PATH" "misc/inkscape_extension/*")))))
         (if (or (os-win32?) (os-mingw?)) 
           (system (string-append "copy " source-dir " " dest-dir))
           (system (string-append "cp " source-dir " " dest-dir)))
       (refresh-now "svg-configutation")))

(define (get-equation-editor-server)
  (get-boolean-preference "equation-editor-server"))

(define (set-equation-editor-server on?)
  (set-boolean-preference "equation-editor-server" on?)
  (refresh-now "svg-configutation"))

(define (supports-inkscape?) (url-exists-in-path? "inkscape"))

(tm-widget (image-preferences-widget)
  ===
  (bold (text "Image -> TeXmacs"))
  ===
  (aligned
    (meti
      (when (supports-inkscape?)
        (hlist // (text "Use Inkscape for conversion from SVG")))
      (when (supports-inkscape?)
        (toggle (set-boolean-preference
                 "image->texmacs:svg-prefer-inkscape" answer)
                (get-boolean-preference
                 "image->texmacs:svg-prefer-inkscape")))))
  ===
  (bold (text "TeXmacs -> Image"))
  ===
  (refreshable "svg-configutation"
    (aligned
      (item (text "Bitmap resolution (dpi):")
        (enum (set-preference "texmacs->image:raster-resolution" answer)
	      '("1200" "600" "300" "150" "")
	      (get-preference "texmacs->image:raster-resolution")
	      "5em"))
      (item (text "Clipboard image format:")
        (enum (set-pretty-preference "texmacs->image:format" answer)
	      (pretty-format-list)
	      (get-pretty-preference "texmacs->image:format")
          "5em"))
	  (item (text "Output scale factor:")
        (enum (set-pretty-preference "texmacs->image:scale" answer)
          '("1" "1.5" "2" "")
          (get-pretty-preference "texmacs->image:scale")
          "5em"))) 
    
    ======
  (bold (text "Configure TeXmacs as equation editor external programs"))
  ===
    (aligned (meti (hlist  // (text "Svg converter available") >>> 
                (if (not (svg-converter-available)) 
				  (explicit-buttons ("get converter" (get-svg-converter)))))
          (inert (toggle (noop) (svg-converter-available)))))
	;; tests in "assuming" not refreshed, using clumsy "if" construct instead	  
    (if (not (or (os-win32?) (os-mingw?))) 
	  (aligned (meti (hlist  // (text "Inkscape installed"))
          (inert (toggle (noop) (inkscape-detected))))))
    (if (and (inkscape-detected) (not (inkscape-prefs-found)))
      (aligned (meti (hlist  // (text "Inkscape user prefs found") >>> 
          (if (not (inkscape-prefs-found))
            (explicit-buttons (
			  (eval (if (or (os-win32?) (os-mingw?)) "help" "create (start Inkscape)"))
                (start-inkscape))) ))
          (inert (toggle (noop) (inkscape-prefs-found))))))
    (if (and (inkscape-detected) (inkscape-prefs-found) (svg-converter-available))
      (aligned (meti (hlist  // (text "Inkscape extension installed") >>> 
		    (explicit-buttons ("install/update extension" (install-inkscape-extension))))              
            (inert (toggle (noop) (inkscape-extension-installed))))
         (meti (hlist  // (text "Enable socket communications (faster)"))              
            (toggle (set-equation-editor-server answer)
              (get-equation-editor-server))))))

    ======
      (centered
       (hlist  /// (explicit-buttons ("Refresh" (refresh-now "svg-configutation")))
           /// (explicit-buttons ("Help" (load-help-buffer (url-resolve (url-glue (string->url "$TEXMACS_DOC_PATH") "/main/convert/man-graphics-export.en.tm") "r"))))))

)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; private functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (baseline-relpos t)
;; compute relative position of baseline
  (number->string (exact->inexact
    (/ 
      (- 
        (string->number(cadr (box-info (stree->tree `(with "text-at-valign" "bottom" ,t)) "B")))
        (string->number(cadr (box-info (stree->tree `(with "text-at-valign" "base" ,t)) "B"))))
 ;       (string->number(cadr (box-info (stree->tree `(with "doc-at-valign" "bottom" ,t)) "B")))
  ;      (string->number(cadr (box-info (stree->tree `(with "doc-at-valign" "base" ,t)) "B"))))
      (string->number(cadr (box-info t "H")))))))

(define (url-temp-ext ext)
;; temporary files with extensions
  (url-glue (url-temp) (string-append "." ext)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; commodity functions for tree manipulations
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (remove-node! node)
  ;; removes node (with children if any)
  (tree-remove! (tree-ref node :up) (tree-index node) 1))

(define (copy-node! node parent-dest pos)
  ;; insert an existing node (with children if any) as new child of parent-dest
  ;; FIXME: No sanity check! parent should not be in node's subtree!
  (tree-insert! parent-dest pos `(,node)))

(define (move-node! node parent-dest pos)
  ;; moves an existing node (with children if any) and
  ;; insert it as new child of parent-dest
  ;; FIXME: No sanity check! parent should not be in node's subtree!
  (copy-node! node parent-dest pos)
  (remove-node! node))

(define (remove-node-raise-children! node . firstlast)
  ;; similar to "remove tag" operation in the editor
  (let* ((parent (tree-ref node :up))
         (pos (+ (tree-index node) 1))
         (lastindex (- (if (< (length firstlast) 2) (tree-arity node) (cadr firstlast) )1))
         (firstindex (- (if (null? firstlast) 1 (car firstlast)) 1)))
;;    (display* firstindex " " lastindex "\n")  
    (do ((i lastindex (- i 1))) ((= i firstindex))
      (copy-node! (tree-ref node i) parent pos))
    (remove-node! node)
    ))

(define (replace-leaftext! leaf newtext)
  ;; replace a node's content by a new string.
  ;; Makes the node a leaf if it wasn't one
  (tree-assign! leaf (string->tree newtext)))

(define (selection-trim-ending)
 	  	   (if (selection-active-any?)
	  	     (with st (selection-tree)
 	  	       (if (and (not  (tree-atomic? st ))
	  	                (tree-empty? (tree-ref st :last)))
	  	         (begin
	  	           (selection-set
 	  	             (selection-get-start)
 	  	             (path-previous (root-tree) (selection-get-end)))
	  	           (selection-trim-ending))))))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main functions needed for making reeditable svg
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (replace-hlinks! svgroot defs tm-fragment-string)
;; function for remapping cross-referenced items (glyphs)
;; in the svg using unique ids. This is needed to avoid collisions between
;; definitions belonging to differents formulas in inkscape
  (let* 
    ((deflist (select defs '(:* @ id :%1)))
     ;; list of all ids in the drawing, pointed to by 'clip-path'
     ;; attributes or xlink:href attribute of 'use' nodes 
     (hreflist (select svgroot '(:* @ (:or xlink:href clip-path style) :%1)))
     (unique (number->string (string-hash tm-fragment-string 1000000)))
     ;; generate a reproducible 6-digit number that depends
     ;; on the tm code of the selection
     (basename (string-append "tm" unique "-"))
     (n (length deflist)))
    (if (> n 0)
      (do ((i 0 (+ i 1))) ((= i n))
        (let* 
          ((labelnode (list-ref deflist i))
           (deftype (tree-label (tree-ref labelnode :up :up :up)))
           (newlabel (string-append basename (number->string i)))
           (callerfound #f))
          (map 
            (lambda (x) 
              (let* 
                ((stylestr (tree->string x))
                 (labelstr (tree->string labelnode))
                 (labelength (string-length labelstr))
                 (pos (string-contains stylestr labelstr)))
                (if (and pos (or (== (string-length stylestr) (+ pos labelength)) (== (substring stylestr (+ pos labelength) (+ 1 pos labelength)) ")")))
                  (begin 
                    (replace-leaftext! x (string-replace stylestr labelstr newlabel))
                    (set! callerfound #t)))))
            hreflist)
          (if callerfound
            (replace-leaftext! labelnode newlabel)
            (remove-node! (tree-ref labelnode :up :up :up))) ;;that def was previously removed or never used (it happens!)
        )))))

;; the svg generated by poppler from texmacs pdf output
;; has many useless <g clip-path=xxx some_element> (and even often nested)
;; making the svg unecessarily complicated and big.
;; We remove unecessary clip-pathes and flatten some structures.
;; The resulting svg for simple text and formulas is clean
;; WARNING : these functions could cause errors if the pdf->svg converter
;; is not poppler-based (pdftocairo or pdf2svg)

(define (clean-svg1! svgroot)
;remove useless outer clip-path in nested g clip-path g, but keep id if exists
;oct 2017: disable because messes up streched braces
  (let* 
    ((gclipg (reverse(select svgroot '(:* g @ clip-path :up :up g :up)))))
    (map-in-order 
      (lambda (x) 
        (let* ((childg (tree-ref x 'g)) 
               (gattr (tree-ref x '@)) 
               (gid (tree-ref gattr 'id))) 
          (if (nnot gid) (move-node! (tree-ref gattr 'id) (tree-ref childg '@) 1))
          (remove-node! gattr )
          (remove-node-raise-children! x))
          )
       gclipg)))

(define (clean-svg2! maingroup)
;;in main group flatten g use, keeping g attributes except clip-paths
  (let* 
    ((guse (reverse (select maingroup '(g use)))))
    (map-in-order 
      (lambda (x) 
        (let* ((parentg (tree-ref x :up))) 
          (map remove-node! (select parentg '(@ (:or clip-rule clip-path))))
          (map (lambda (y) (move-node! y (tree-ref x '@) 1)) (select parentg '(@ :%1)))
          (remove-node! (tree-ref parentg '@))
          (remove-node-raise-children! parentg)))
      guse)))

(define (clean-svg3! defs)
;;in defs flatten (g symbol path), keeping symbol attributes
  (let* 
    ((gsymbolpath (reverse (select defs '(g symbol path :up))))
     (gsymbol #f))
    (if (nnull? gsymbolpath)
      (begin
        (set! gsymbol (tree-ref (car gsymbolpath) :up))
        (map-in-order 
          (lambda (x) 
            (let* ((symbolpath (tree-ref x 'path))) 
            (map (lambda (y) (move-node! y (tree-ref symbolpath '@) 1)) (select x '(@ :%1)))
            (remove-node! (tree-ref x '@))
            (remove-node-raise-children! x)))
          gsymbolpath)
        (remove-node-raise-children! gsymbol)))))

;; define latex and texmacs string representation of selection
;; we escape them to ascii so that they do not interfere with xml
;; <  -> &lt;  > -> &gt; \ -> \\, all characters above #127->\xXX ...
;; see  TeXmacs/langs/encodings/cork-escaped-to-ascii.scm
(define (latex-encode tm-fragment-tree)
  ;; for the latex representation we mimick what is done when
  ;; "copy to latex" is performed
  (let* ((latex-tree (latex-expand tm-fragment-tree))
         ;; expand or not macros according to preferences
         (latex-code (texmacs->generic latex-tree "latex-snippet")))
         ;; actual conversion
    (escape-to-ascii latex-code)))

(define (tm-encode tm-fragment-tree)
  (escape-to-ascii (serialize-texmacs tm-fragment-tree)))

(define (refactor-svg dest tm-fragment relbaseline)
;; dest is the url of the svg file to be edited
;; We reorganize & cleanup the svg file and inject attributes containing:
;; - the tm code of equation
;; - style info from the original document (style, fonts, layout, ...)
;; - A latex fragment for compatibility with 'textext' inkscape extension
;; - the relative position of the baseline to enable vertical alignement
;;   in an external application
;; FIXME : no error checking, no return value...

  (let*
    ( ;; 1: load svg and transform to an active tree in
      ;; temporary buffer so that we can manipulate it
      ;; using texmacs primitives for trees
      (svg-in (string-load dest)) ;; load svg file as string
      (s-svg-in (parse-xml svg-in)) ;; parse to stree
      (mybuf (buffer-new))
      ;; create temporary buffer for subsequent manipulations of svg tree
      (void (buffer-set-body mybuf (tree-assign-node! (stree->tree s-svg-in)
                                                     'concat)))
      ;; populate buffer with tree
      ;; replace *TOP* node by concat otherwise displaying
      ;; that buffer crashes texmacs
      
      ;; 2: define a bunch of locations in the tree
      (buftree (buffer-get-body mybuf)) ;; the whole tree
      (svgroot (car (select buftree '(:* svg)))) ;; the <svg > node
      (maingroup (car (select svgroot '(g))))
      ;; the main group in the svg, containing the drawing layout
      (maingroup-attrib (car (select maingroup '(@))))
      ;; attributes of the main group
      (defs (car (select svgroot '(defs))))
      ;; the defs, containing the glyph vector outlines,
      ;; hyperlinked from the drawing (a.k.a cloned)
      (bgframe (select maingroup '(:* path @ style :%1)))
      (bgframe (if (null? bgframe) #f (car bgframe)))
      ;; the solid background we introduced  
      ;; 3: the new data we want to insert in the tree
      (latex-code (latex-encode tm-fragment))
      (tm-code (tm-encode tm-fragment))
      (tm-style (tm-encode (get-all-inits)))
      ;; define new attributes containing latex and texmacs code:
      (extra-latex-attrib
      `((xmlns:ns0 "http://www.iki.fi/pav/software/textext/")
        (ns0:text ,latex-code) (ns0:preamble "texmacs_latex.sty")))
      (extra-tm-attrib 
      `((xmlns:ns1 "https://www.texmacs.org/")
        (ns1:texmacscode ,tm-code) (ns1:texmacsstyle ,tm-style)
        (ns1:texmacsbaseline ,relbaseline)))
      ;; OK, the texmacs namespace maybe not correctly described at that url ...
    )

    ;; 4: modify tree
    ;; 4.0 set the background fully transparent
    (if (and bgframe (== (tree->string bgframe) " stroke:none;fill-rule:nonzero;fill:rgb(100%,100%,100%);fill-opacity:0.007;")) 
        (tree-set! bgframe " stroke:none;fill-rule:nonzero;fill:rgb(100%,100%,100%);fill-opacity:0.00"))
    ;; 4.1, (optional) simplify svg tree, removing unecessary clips and structures
    ;; we don't even try simplifiying if the selection contains an image
    ;; as it may have an arbitrarily complex svg structure
    (if (not (or (== (tree-label tm-fragment) 'image) (nnull? (select tm-fragment '(:* image))))) 
      (begin
        ;;(clean-svg1! svgroot) ;oct 2017: disable because messes up streched braces
        (clean-svg2! maingroup)
        (clean-svg3! defs)))
    ;; 4.2, replace hlinks with new pointers, removing unused defs
    (replace-hlinks! svgroot defs (string-append tm-code tm-style (get-preference "texmacs->image:scale")))
    ;; 4.3 (not optional!), add our own new attributes for re-editting equation  
    (tree-insert! maingroup-attrib 1 extra-latex-attrib) ;; for textext compatibility
    (tree-insert! maingroup-attrib 2 extra-tm-attrib)
    ;; 4.4 move defs containing the glyph outlines inside main group
    ;; so that they remain together in inkscape (doesn't work when we go through clipboard, though)
    (move-node! defs maingroup 2)

    ;; 5: finally create output
    (let* (;; convert back to stree, recreate the *TOP* node,
           ;; and restore *PI* xml
           ;; (instead of *PI* "xml" given by tree->stree -
           ;; otherwise serialize-html fails)
           (s-svg-out
            (append '(*TOP* (*PI* xml "version=\"1.0\" encoding=\"UTF-8\""))
                    ;; actually we use only ascii
                    (cddr (tree->stree buftree))))
           (xml-svg-out (begin (output-flush) ;; necessary??
                               (serialize-tmml s-svg-out))))
      ;; close temporary buffer
      (buffer-pretend-saved mybuf)
      (buffer-close mybuf)
      (string-save xml-svg-out dest))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; public interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tm-define (clipboard-copy-image void)
  (:synopsis "Places an image of the current selection on the clipboard")
  (:argument void "not used")
  (:returns "nothing")
  ;;the format of the graphics is set in the preferences
  (if (not (qt-gui?))
    (set-message "Qt GUI only, sorry. Use \"Export selection...\"" "")
    (if (not (selection-active-any?))
      (set-message "no selection!" "")
      (begin (selection-trim-ending)
        (let* ((format (get-preference "texmacs->image:format"))
             (tmpurl (url-temp-ext format)))
          (export-selection-as-graphics tmpurl)
	;; first generate an image file
          (graphics-file-to-clipboard tmpurl)
	;; place that image on the clipboard
          (system-remove tmpurl)
	)))))

(tm-define (export-selection-as-graphics myurl)
  (:synopsis "Generates graphics format of the current selection")
  (:argument myurl "A full file url with extension")
  (:returns "nothing")
  ;; for svg export, the texmacs code of the selection as well as 
  ;; global document parameters such as style, fonts, etc. are
  ;; embedded in the drawing to enable re-editing from inkscape
 
  (cond
   ((not (selection-active-any?)) (set-message "no selection!" ""))
   ((not (file-converter-exists? "x.pdf" (string-append "y." (url-suffix myurl)) )) 
      (show-message (string-append "pdf to " (url-suffix myurl) " converter not available. Sorry.") "Cannot generate image" ))
   (else 
    (let*
     (
;; step 1 prepare and typeset selection
;; if selection is inside inline or display math preserve inline/display style

      (tm-fragment
        (cond
          ((match? (tree->stree (selection-tree)) 
                    '(:or (equation* :*) (equation :*) (eqnarray :*) (eqnarray* :*) (math :*)))
            (display "is some math \n") (selection-tree)) 
          ((==  (tree->string (get-env-tree-at "mode" (selection-get-start))) "math")
             (if (==  (tree->string (get-env-tree-at "math-display" (selection-get-start))) "true")
               (begin 
                 (display "in display math \n") 
                 (stree->tree `(equation* (document ,(selection-tree)))))
               (begin 
                 (display "in math \n")
	             (stree->tree `(math (document ,(selection-tree)))))))
          (else (display "other \n")(selection-tree))))
;; is selection wider than 1par (and needs linebreaks and or hyphenation)?
      (maxwidth (length-decode "1par"))
      (partmpt (string-append (number->string maxwidth) "tmpt"))
      (parcm (length-add  "0cm" partmpt))
      (pagewcm (length-add  "0cm" (string-append (number->string (+ 1 (*  (get-page-width #t) 10))) "tmpt")))
      ;; get-page-width return dimension in the cpp "SI" unit which seems to be == "10tmpt"
;; output scale of image?
      (str-scale  (get-preference "texmacs->image:scale"))
      (scale (string->number str-scale))

;; We compute the baseline position only if it's a single-line content
;; This excludes selections begining with 'document
;; If the selection is a 'concat we need to check if it does not exceed the page width
;; If the selection is eqnarray(*) compute it only if it has a single line
;;  (table-nr-rows) works for this!
      (iseqnarray (match? (tree->stree tm-fragment) '(:or (eqnarray :*) (eqnarray* :*) (:* align* :*))))
      (simpleeqnarray (and iseqnarray (== 1 (table-nr-rows)))) 
      (needbaseline 
        (if (match? (tree->stree tm-fragment) '(document :*)) 
          #f   
         (or
          simpleeqnarray
          (< (string->number (cadr (box-info tm-fragment "W"))) (length-decode "1par")))))
;; the baseline calculation is relative to the size of the background
;; if not fully transparent the background is included in the svg image (otherwise no)
;; (during svg postprocessing  we'll set the opacity of the background to 0)  
      (fillcolor (if needbaseline "#ffffff02" "#ffffff00")) ; either slightly opaque or fully transparent white


;; if needbaseline use text-at 
;; otherwise (multiline selection) use doc-at to get proper pagewidth
      (tm-fragment-formated 
         (if iseqnarray 
           `(document (tformat (twith table-hmode min) ,tm-fragment)) ;reduce table width to min, to avoid wide empty frame
           tm-fragment))
      (tm-fragment-formated
        (if needbaseline
          (stree->tree `(with "magnify" ,str-scale "page-width" 
            ,(length-mult scale pagewcm)
            (text-at 
             (tabular (tformat 
                (twith "table-width" ,(length-mult scale parcm))
                (twith "table-hmode" "min")
                (twith "table-valign" "B") ;;(=top baseline)
                (cwith "1" "-1" "1" "-1" "cell-background" ,fillcolor)
                (cwith "1" "1" "1" "1" "cell-hyphen" "t")
                (cwith "1" "-1" "1" "-1" "cell-lsep" "0spc")
                (cwith "1" "-1" "1" "-1" "cell-tsep" "0sep")
                (cwith "1" "-1" "1" "-1" "cell-rsep" "0spc")
                (cwith "1" "-1" "1" "-1" "cell-bsep" "0sep")
                (table (row (cell (document ,tm-fragment))))))
              (point "0par" "0"))))
        (stree->tree `(with "magnify" ,str-scale "page-width" 
          ,(length-mult scale pagewcm) 
          "doc-at-width" ,(length-mult scale parcm) "doc-at-hmode" "min" "doc-at-padding" "0spc";"doc-at-ppsep" "" 
             (document-at (document ,tm-fragment-formated ) (point "0par" "0"))))))

;; compute relative position of baseline  
      (relbaseline (if needbaseline (baseline-relpos tm-fragment-formated) "0.0"))
      (tm-fragment-formated
        (if needbaseline 
           `(with "text-at-valign" "base" ,tm-fragment-formated)
           tm-fragment-formated))

    (temp1 (url-temp-ext "pdf"))
    (suffix (url-suffix myurl)))

    (print-snippet temp1 (stree->tree tm-fragment-formated) #t)

	;; step 2 generate output according to desired output format

	  (cond 
	    ((== suffix "pdf") (system-copy temp1 myurl))
	    ((== suffix "svg")
	       ;; assume target is inkscape
		   (file-convert temp1 myurl)
		 ;; using either pdf2svg or pdf2cairo converters is crucial
		 ;; for svg inport in inkscape:
		 ;; fonts are properly passed as vector outlines
		 ;; file converters are defined in init-images.scm
         (display* "relbaseline= " relbaseline "\n")
         (refactor-svg myurl tm-fragment relbaseline))
		 ;; modify svg, embedding texmacs code
	    (else
            (file-convert temp1 myurl)))

	  (system-remove temp1) ;; temp pdf file not needed anymore
	))))

